#!/usr/bin/env python
# This is the example used in the documentation.

SERIAL_PORT_PATH="/dev/cu.usbmodem14434401"  # your actual path to the Arduino Native serial port device goes here
import sys

import hackeeg
from hackeeg import ads1299

hackeeg = hackeeg.HackEEGBoard(SERIAL_PORT_PATH)
hackeeg.connect()
hackeeg.sdatac()
hackeeg.reset()
hackeeg.blink_board_led()
hackeeg.disable_all_channels()
sample_mode = ads1299.HIGH_RES_250_SPS | ads1299.CONFIG1_const
hackeeg.wreg(ads1299.CONFIG1, sample_mode)
test_signal_mode = ads1299.INT_TEST_4HZ | ads1299.CONFIG2_const
hackeeg.wreg(ads1299.CONFIG2, test_signal_mode)
hackeeg.enable_channel(7)
hackeeg.wreg(ads1299.CH7SET, ads1299.TEST_SIGNAL | ads1299.GAIN_1X)
hackeeg.rreg(ads1299.CH5SET)

# Unipolar mode - setting SRB1 bit sends mid-supply voltage to the N inputs
hackeeg.wreg(ads1299.MISC1, ads1299.SRB1)
# add channels into bias generation
hackeeg.wreg(ads1299.BIAS_SENSP, ads1299.BIAS8P)
hackeeg.rdatac()
hackeeg.start()

while True:
    result = hackeeg.read_response()
    status_code = result.get('STATUS_CODE')
    status_text = result.get('STATUS_TEXT')
    data = result.get(hackeeg.DataKey)
    if data:
        decoded_data = result.get(hackeeg.DecodedDataKey)
        if decoded_data:
            timestamp = decoded_data.get('timestamp')
            ads_gpio = decoded_data.get('ads_gpio')
            loff_statp = decoded_data.get('loff_statp')
            loff_statn = decoded_data.get('loff_statn')
            channel_data = decoded_data.get('channel_data')
            print(f"timestamp:{timestamp} | gpio:{ads_gpio} loff_statp:{loff_statp} loff_statn:{loff_statn} |   ",
                  end='')
            for channel_number, sample in enumerate(channel_data):
                print(f"{channel_number + 1}:{sample} ", end='')
            print()
        else:
            print(data)
        sys.stdout.flush()




import base64
import binascii
import io
import json
import sys
from json import JSONDecodeError

import msgpack
import serial
import time

from . import ads1299

# TODO
# - MessagePack
# - MessagePack / Json Lines testing convenience functions

NUMBER_OF_SAMPLES = 10000
DEFAULT_BAUDRATE = 115200
SAMPLE_LENGTH_IN_BYTES = 38  # 216 bits encoded with base64 + '\r\n\'

SPEEDS = {250: ads1299.HIGH_RES_250_SPS,
          500: ads1299.HIGH_RES_500_SPS,
          1000: ads1299.HIGH_RES_1k_SPS,
          2000: ads1299.HIGH_RES_2k_SPS,
          4000: ads1299.HIGH_RES_4k_SPS,
          8000: ads1299.HIGH_RES_8k_SPS,
          16000: ads1299.HIGH_RES_16k_SPS}

GAINS = {1: ads1299.GAIN_1X,
         2: ads1299.GAIN_2X,
         4: ads1299.GAIN_4X,
         6: ads1299.GAIN_6X,
         8: ads1299.GAIN_8X,
         12: ads1299.GAIN_12X,
         24: ads1299.GAIN_24X}


class Status:
    Ok = 200
    BadRequest = 400
    Error = 500


class HackEEGException(Exception):
    pass


class HackEEGBoard:
    TextMode = 0
    JsonLinesMode = 1
    MessagePackMode = 2

    CommandKey = "COMMAND"
    ParametersKey = "PARAMETERS"
    HeadersKey = "HEADERS"
    DataKey = "DATA"
    DecodedDataKey = "DECODED_DATA"
    StatusCodeKey = "STATUS_CODE"
    StatusTextKey = "STATUS_TEXT"

    MpCommandKey = "C"
    MpParametersKey = "P"
    MpHeadersKey = "H"
    MpDataKey = "D"
    MpStatusCodeKey = "C"
    MpStatusTextKey = "T"
    MaxConnectionAttempts = 10
    ConnectionSleepTime = 0.1

    def __init__(self, serial_port_path=None, baudrate=DEFAULT_BAUDRATE, debug=False):
        self.mode = None
        self.message_pack_unpacker = None
        self.debug = debug
        self.baudrate = baudrate
        self.rdatac_mode = False
        self.serial_port_path = serial_port_path
        if serial_port_path:
            self.raw_serial_port = serial.serial_for_url(serial_port_path, baudrate=self.baudrate, timeout=0.1)
            self.raw_serial_port.reset_input_buffer()
            # self.serial_port= self.raw_serial_port
            self.serial_port = io.TextIOWrapper(io.BufferedRWPair(self.raw_serial_port, self.raw_serial_port))
            # self.binaryBufferedSerialPort = io.BufferedReader(io.BufferedRWPair(self.raw_serial_port, self.raw_serial_port))
            # self.message_pack_unpacker = msgpack.Unpacker(self.binaryBufferedSerialPort, raw=False, use_list=False)
            self.message_pack_unpacker = msgpack.Unpacker(self.raw_serial_port, raw=False, use_list=False)

    def connect(self):
        self.mode = self._sense_protocol_mode()
        if self.mode == self.TextMode:
            attempts = 0
            connected = False
            while attempts < self.MaxConnectionAttempts:
                try:
                    self.jsonlines_mode()
                    connected = True
                    break
                except JSONDecodeError:
                    if attempts == 0:
                        print("Connecting...", end='')
                    elif attempts > 0:
                        print('.', end='')
                    sys.stdout.flush()
                    attempts += 1
                    time.sleep(self.ConnectionSleepTime)
            if attempts > 0:
                print()
            if not connected:
                raise HackEEGException("Can't connect to Arduino")
        self.sdatac()
        line = self.serial_port.readline()
        while line:
            line = self.serial_port.readline()

    def _serial_write(self, command):
        self.serial_port.write(command)
        self.serial_port.flush()

    def _serial_readline(self, serial_port=None):
        if serial_port is None:
            line = self.serial_port.readline()
        elif serial_port == "raw":
            line = self.raw_serial_port.readline()
        else:
            raise HackEEGException('Unknown serial port designator; must be either None or "raw"')
        return line

    def _serial_read_messagepack_message(self):
        message = self.message_pack_unpacker.unpack()
        if self.debug:
            print(f"message: {message}")
        return message

    def _decode_data(self, response):
        """decode ADS1299 sample status bits - datasheet, p36
        The format is:
        1100 + LOFF_STATP[0:7] + LOFF_STATN[0:7] + bits[4:7] of the GPIOregister"""
        error = False
        if response:
            data = response.get(self.DataKey)
            if data is None:
                data = response.get(self.MpDataKey)
                if type(data) is str:
                    try:
                        data = base64.b64decode(data)
                    except binascii.Error:
                        print(f"incorrect padding: {data}")

            if data and (type(data) is list or type(data) is bytes):
                data_hex = ":".join("{:02x}".format(c) for c in data)
                if error:
                    print(data_hex)
                timestamp = int.from_bytes(data[0:4], byteorder='little')
                sample_number = int.from_bytes(data[4:8], byteorder='little')
                ads_status = int.from_bytes(data[8:11], byteorder='big')
                ads_gpio = ads_status & 0x0f
                loff_statn = (ads_status >> 4) & 0xff
                loff_statp = (ads_status >> 12) & 0xff
                extra = (ads_status >> 20) & 0xff

                channel_data = []
                for channel in range(0, 8):
                    channel_offset = 11 + (channel * 3)
                    sample = int.from_bytes(data[channel_offset:channel_offset + 3], byteorder='big', signed=True)
                    channel_data.append(sample)

                response['timestamp'] = timestamp
                response['sample_number'] = sample_number
                response['ads_status'] = ads_status
                response['ads_gpio'] = ads_gpio
                response['loff_statn'] = loff_statn
                response['loff_statp'] = loff_statp
                response['extra'] = extra
                response['channel_data'] = channel_data
                response['data_hex'] = data_hex
                response['data_raw'] = data
        return response

    def set_debug(self, debug):
        self.debug = debug

    def read_response(self, serial_port=None):
        """read a response from the Arduino– must be in JSON Lines mode"""
        message = self._serial_readline(serial_port=serial_port)
        try:
            response_obj = json.loads(message)
        except UnicodeDecodeError:
            response_obj = None
        # except JSONDecodeError:
        #     response_obj = None
        if self.debug:
            print(f"read_response line: {message}")
        if self.debug:
            print("json response:")
            print(self.format_json(response_obj))
        return self._decode_data(response_obj)

    def read_rdatac_response(self):
        """read a response from the Arduino– JSON Lines or MessagePack mode are ok"""
        if self.mode == self.MessagePackMode:
            response_obj = self._serial_read_messagepack_message()
        else:
            message = self._serial_readline()
            try:
                response_obj = json.loads(message)
            except JSONDecodeError:
                response_obj = {}
                print()
                print(f"json decode error: {message}")
        if self.debug:
            print(f"read_response obj: {response_obj}")
        result = None
        try:
            result = self._decode_data(response_obj)
        except AttributeError:
            pass
        return result

    def format_json(self, json_obj):
        return json.dumps(json_obj, indent=4, sort_keys=True)

    def send_command(self, command, parameters=None):
        if self.debug:
            print(f"command: {command}  parameters: {parameters}")
        # commands are only sent in JSON Lines mode
        new_command_obj = {self.CommandKey: command, self.ParametersKey: parameters}
        new_command = json.dumps(new_command_obj)
        if self.debug:
            print("json command:")
            print(self.format_json(new_command_obj))
        self._serial_write(new_command)
        self._serial_write('\n')

    def send_text_command(self, command):
        self._serial_write(command + '\n')

    def execute_command(self, command, parameters=None, serial_port=None):
        if parameters is None:
            parameters = []
        self.send_command(command, parameters)
        response = self.read_response(serial_port=serial_port)
        return response

    def _sense_protocol_mode(self):
        try:
            self.send_command("stop")
            self.send_command("sdatac")
            result = self.execute_command("nop")
            return self.JsonLinesMode
        except Exception:
            return self.TextMode

    def ok(self, response):
        return response.get(self.StatusCodeKey) == Status.Ok

    def wreg(self, register, value):
        command = "wreg"
        parameters = [register, value]
        return self.execute_command(command, parameters)

    def rreg(self, register):
        command = "rreg"
        parameters = [register]
        response = self.execute_command(command, parameters)
        return response

    def nop(self):
        return self.execute_command("nop")

    def boardledon(self):
        return self.execute_command("boardledon")

    def boardledoff(self):
        return self.execute_command("boardledoff")

    def ledon(self):
        return self.execute_command("ledon")

    def ledoff(self):
        return self.execute_command("ledoff")

    def micros(self):
        return self.execute_command("micros")

    def text_mode(self):
        return self.send_command("text")

    def reset(self):
        return self.execute_command("reset")

    def start(self):
        return self.execute_command("start")

    def stop(self):
        return self.execute_command("stop")

    def rdata(self):
        return self.execute_command("rdata")

    def version(self):
        result = self.execute_command("version")
        return result

    def status(self):
        return self.execute_command("status")

    def jsonlines_mode(self):
        old_mode = self.mode
        self.mode = self.JsonLinesMode
        if old_mode == self.TextMode:
            self.send_text_command("jsonlines")
            return self.read_response()
        if old_mode == self.JsonLinesMode:
            self.execute_command("jsonlines")

    def messagepack_mode(self):
        old_mode = self.mode
        self.mode = self.MessagePackMode
        if old_mode == self.TextMode:
            self.send_text_command("jsonlines")
            response = self.read_response()
            self.execute_command("messagepack")
            return response
        elif old_mode == self.JsonLinesMode:
            response = self.execute_command("messagepack")
            return response

    def rdatac(self):
        result = self.execute_command("rdatac", serial_port="raw")
        if self.ok(result):
            self.rdatac_mode = True
        return result

    def sdatac(self):
        if self.mode == self.JsonLinesMode:
            result = self.execute_command("sdatac")
        else:
            self.send_command("sdatac")
            result = self.read_response(serial_port="raw")
        self.rdatac_mode = False
        return result

    def stop_and_sdatac_messagepack(self):
        """used to smoothly stop data transmission while in MessagePack mode–
        mostly avoids exceptions and other hiccups"""
        self.send_command("stop")
        self.send_command("sdatac")
        self.send_command("nop")
        try:
            line = self.serial_port.read()
        except UnicodeDecodeError:
            line = self.raw_serial_port.read()

    def enable_channel(self, channel, gain=None):
        if gain is None:
            gain = ads1299.GAIN_1X
        temp_rdatac_mode = self.rdatac_mode
        if self.rdatac_mode:
            self.sdatac()
        command = "wreg"
        parameters = [ads1299.CHnSET + channel, ads1299.ELECTRODE_INPUT | gain]
        self.execute_command(command, parameters)
        if temp_rdatac_mode:
            self.rdatac()

    def disable_channel(self, channel):
        command = "wreg"
        parameters = [ads1299.CHnSET + channel, ads1299.PDn | ads1299.SHORTED]
        self.execute_command(command, parameters)

    def enable_all_channels(self):
        for channel in range(1, 9):
            self.enable_channel(channel)

    def disable_all_channels(self):
        for channel in range(1, 9):
            self.disable_channel(channel)

    def blink_board_led(self):
        self.execute_command("boardledon")
        time.sleep(0.3)
        self.execute_command("boardledoff")



# ADS1299 constants

# system commands
WAKEUP = 0x02
STANDBY = 0x04
RESET = 0x06
START = 0x08
STOP = 0x0a

# read commands
RDATAC = 0x10
SDATAC = 0x11
RDATA = 0x12

# register commands
RREG = 0x20
WREG = 0x40

# device settings
ID = 0x00

# global settings
CONFIG1 = 0x01
CONFIG2 = 0x02
CONFIG3 = 0x03
LOFF = 0x04

# channel specific settings
CHnSET = 0x04
CH1SET = CHnSET + 1
CH2SET = CHnSET + 2
CH3SET = CHnSET + 3
CH4SET = CHnSET + 4
CH5SET = CHnSET + 5
CH6SET = CHnSET + 6
CH7SET = CHnSET + 7
CH8SET = CHnSET + 8
BIAS_SENSP = 0x0d
BIAS_SENSN = 0x0e
LOFF_SENSP = 0x0f
LOFF_SENSN = 0x10
LOFF_FLIP = 0x11

# lead off status
LOFF_STATP = 0x12
LOFF_STATN = 0x13

GPIO = 0x14
MISC1 = 0x15
RESP = 0x16
CONFIG4 = 0x17
WCT1 = 0x18
WCT2 = 0x19

DEV_ID7 = 0x80
DEV_ID6 = 0x40
DEV_ID5 = 0x20
DEV_ID3 = 0x08
DEV_ID2 = 0x04
DEV_ID1 = 0x02
DEV_ID0 = 0x01

ID_const = 0x10
ID_ADS129x = DEV_ID7
ID_ADS129xR = (DEV_ID7 | DEV_ID6)

ID_4CHAN = 0
ID_6CHAN = DEV_ID0
ID_8CHAN = DEV_ID1

ID_ADS1294 = (ID_ADS129x | ID_4CHAN)
ID_ADS1296 = (ID_ADS129x | ID_6CHAN)
ID_ADS1298 = (ID_ADS129x | ID_8CHAN)
ID_ADS1294R = (ID_ADS129xR | ID_4CHAN)
ID_ADS1296R = (ID_ADS129xR | ID_6CHAN)
ID_ADS1298R = (ID_ADS129xR | ID_8CHAN)
ID_ADS1299 = (DEV_ID3 | DEV_ID2 | DEV_ID1)

HR = 0x80
DAISY_EN = 0x40
CLK_EN = 0x20
DR2 = 0x04
DR1 = 0x02
DR0 = 0x01

CONFIG1_const = 0x90
HIGH_RES_16k_SPS = 0x00
HIGH_RES_8k_SPS = (DR0)
HIGH_RES_4k_SPS = (DR1)
HIGH_RES_2k_SPS = (DR1 | DR0)
HIGH_RES_1k_SPS = (DR2)
HIGH_RES_500_SPS = (DR2 | DR0)
HIGH_RES_250_SPS = (DR2 | DR1)

WCT_CHOP = 0x20
INT_TEST = 0x10
TEST_AMP = 0x04
TEST_FREQ1 = 0x02
TEST_FREQ0 = 0x01

CONFIG2_const = 0xC0
INT_TEST_4HZ = INT_TEST
INT_TEST_8HZ = (INT_TEST | TEST_FREQ0)
INT_TEST_DC = (INT_TEST | TEST_FREQ1 | TEST_FREQ0)

PD_REFBUF = 0x80
VREF_4V = 0x20
RLD_MEAS = 0x10
RLDREF_INT = 0x08
PD_RLD = 0x04
RLD_LOFF_SENS = 0x02
RLD_STAT = 0x01

CONFIG3_const = 0x60

COMP_TH2 = 0x80
COMP_TH1 = 0x40
COMP_TH0 = 0x20
VLEAD_OFF_EN = 0x10
ILEAD_OFF1 = 0x08
ILEAD_OFF0 = 0x04
FLEAD_OFF1 = 0x02
FLEAD_OFF0 = 0x01

LOFF_const = 0x00

COMP_TH_95 = 0x00
COMP_TH_92_5 = COMP_TH0
COMP_TH_90 = COMP_TH1
COMP_TH_87_5 = (COMP_TH1 | COMP_TH0)
COMP_TH_85 = COMP_TH2
COMP_TH_80 = (COMP_TH2 | COMP_TH0)
COMP_TH_75 = (COMP_TH2 | COMP_TH1)
COMP_TH_70 = (COMP_TH2 | COMP_TH1 | COMP_TH0)

ILEAD_OFF_6nA = 0x00
ILEAD_OFF_12nA = ILEAD_OFF0
ILEAD_OFF_18nA = ILEAD_OFF1
ILEAD_OFF_24nA = (ILEAD_OFF1 | ILEAD_OFF0)

FLEAD_OFF_AC = FLEAD_OFF0
FLEAD_OFF_DC = (FLEAD_OFF1 | FLEAD_OFF0)

PDn = 0x80
GAINn2 = 0x40
GAINn1 = 0x20
GAINn0 = 0x10
SRB2n0 = 0x08
MUXn2 = 0x04
MUXn1 = 0x02
MUXn0 = 0x01

CHnSET_const = 0x00

# ADS1299
GAIN_1X = 0x00
GAIN_2X = GAINn0
GAIN_4X = GAINn1
GAIN_6X = (GAINn1 | GAINn0)
GAIN_8X = GAINn2
GAIN_12X = (GAINn2 | GAINn0)
GAIN_24X = (GAINn2 | GAINn1)

# ADS1298
ADS1298_GAIN_1X = GAINn0
ADS1298_GAIN_2X = GAINn1
ADS1298_GAIN_3X = (GAINn1 | GAINn0)
ADS1298_GAIN_4X = GAINn2
ADS1298_GAIN_6X = 0x00
ADS1298_GAIN_8X = (GAINn2 | GAINn0)
ADS1298_GAIN_12X = (GAINn2 | GAINn1)

ELECTRODE_INPUT = 0x00
SHORTED = MUXn0
RLD_INPUT = MUXn1
MVDD = (MUXn1 | MUXn0)
TEMP = MUXn2
TEST_SIGNAL = (MUXn2 | MUXn0)
BIAS_DRP = (MUXn2 | MUXn1)
BIAS_DRN = (MUXn2 | MUXn1 | MUXn0)

PD_1 = 0x80
GAIN12 = 0x40
GAIN11 = 0x20
GAIN10 = 0x10
MUX12 = 0x04
MUX11 = 0x02
MUX10 = 0x01

BIAS8P = 0x80
BIAS7P = 0x40
BIAS6P = 0x20
BIAS5P = 0x10
BIAS4P = 0x08
BIAS3P = 0x04
BIAS2P = 0x02
BIAS1P = 0x01

CH1SET_const = 0x00

PD_2 = 0x80
GAIN22 = 0x40
GAIN21 = 0x20
GAIN20 = 0x10
MUX22 = 0x04
MUX21 = 0x02
MUX20 = 0x01

CH2SET_const = 0x00

PD_3 = 0x80
GAIN32 = 0x40
GAIN31 = 0x20
GAIN30 = 0x10
MUX32 = 0x04
MUX31 = 0x02
MUX30 = 0x01

CH3SET_const = 0x00

PD_4 = 0x80
GAIN42 = 0x40
GAIN41 = 0x20
GAIN40 = 0x10
MUX42 = 0x04
MUX41 = 0x02
MUX40 = 0x01

CH4SET_const = 0x00

PD_5 = 0x80
GAIN52 = 0x40
GAIN51 = 0x20
GAIN50 = 0x10
MUX52 = 0x04
MUX51 = 0x02
MUX50 = 0x01

CH5SET_const = 0x00

PD_6 = 0x80
GAIN62 = 0x40
GAIN61 = 0x20
GAIN60 = 0x10
MUX62 = 0x04
MUX61 = 0x02
MUX60 = 0x01

CH6SET_const = 0x00

PD_7 = 0x80
GAIN72 = 0x40
GAIN71 = 0x20
GAIN70 = 0x10
MUX72 = 0x04
MUX71 = 0x02
MUX70 = 0x01

CH7SET_const = 0x00

PD_8 = 0x80
GAIN82 = 0x40
GAIN81 = 0x20
GAIN80 = 0x10
MUX82 = 0x04
MUX81 = 0x02
MUX80 = 0x01

CH8SET_const = 0x00

RLD8P = 0x80
RLD7P = 0x40
RLD6P = 0x20
RLD5P = 0x10
RLD4P = 0x08
RLD3P = 0x04
RLD2P = 0x02
RLD1P = 0x01

RLD_SENSP_const = 0x00

RLD8N = 0x80
RLD7N = 0x40
RLD6N = 0x20
RLD5N = 0x10
RLD4N = 0x08
RLD3N = 0x04
RLD2N = 0x02
RLD1N = 0x01

RLD_SENSN_const = 0x00

LOFF8P = 0x80
LOFF7P = 0x40
LOFF6P = 0x20
LOFF5P = 0x10
LOFF4P = 0x08
LOFF3P = 0x04
LOFF2P = 0x02
LOFF1P = 0x01

LOFF_SENSP_const = 0x00

LOFF8N = 0x80
LOFF7N = 0x40
LOFF6N = 0x20
LOFF5N = 0x10
LOFF4N = 0x08
LOFF3N = 0x04
LOFF2N = 0x02
LOFF1N = 0x01

LOFF_SENSN_const = 0x00

LOFF_FLIP8 = 0x80
LOFF_FLIP7 = 0x40
LOFF_FLIP6 = 0x20
LOFF_FLIP5 = 0x10
LOFF_FLIP4 = 0x08
LOFF_FLIP3 = 0x04
LOFF_FLIP2 = 0x02
LOFF_FLIP1 = 0x01

LOFF_FLIP_const = 0x00

IN8P_OFF = 0x80
IN7P_OFF = 0x40
IN6P_OFF = 0x20
IN5P_OFF = 0x10
IN4P_OFF = 0x08
IN3P_OFF = 0x04
IN2P_OFF = 0x02
IN1P_OFF = 0x01

LOFF_STATP_const = 0x00

IN8N_OFF = 0x80
IN7N_OFF = 0x40
IN6N_OFF = 0x20
IN5N_OFF = 0x10
IN4N_OFF = 0x08
IN3N_OFF = 0x04
IN2N_OFF = 0x02
IN1N_OFF = 0x01

LOFF_STATN_const = 0x00

GPIOD4 = 0x80
GPIOD3 = 0x40
GPIOD2 = 0x20
GPIOD1 = 0x10
GPIOC4 = 0x08
GPIOC3 = 0x04
GPIOC2 = 0x02
GPIOC1 = 0x01

GPIO_const = 0x00

PACEE1 = 0x10
PACEE0 = 0x08
PACEO1 = 0x04
PACEO0 = 0x02
PD_PACE = 0x01

PACE_const = 0x00

PACEE_CHAN2 = 0x00
PACEE_CHAN4 = PACEE0
PACEE_CHAN6 = PACEE1
PACEE_CHAN8 = (PACEE1 | PACEE0)

PACEO_CHAN1 = 0x00
PACEO_CHAN3 = PACEE0
PACEO_CHAN5 = PACEE1
PACEO_CHAN7 = (PACEE1 | PACEE0)

RESP_DEMOD_EN1 = 0x80
RESP_MOD_EN1 = 0x40
RESP_PH2 = 0x10
RESP_PH1 = 0x08
RESP_PH0 = 0x04
RESP_CTRL1 = 0x02
RESP_CTRL0 = 0x01

RESP_const = 0x20

RESP_PH_22_5 = 0x00
RESP_PH_45 = RESP_PH0
RESP_PH_67_5 = RESP_PH1
RESP_PH_90 = (RESP_PH1 | RESP_PH0)
RESP_PH_112_5 = RESP_PH2
RESP_PH_135 = (RESP_PH2 | RESP_PH0)
RESP_PH_157_5 = (RESP_PH2 | RESP_PH1)

RESP_NONE = 0x00
RESP_EXT = RESP_CTRL0
RESP_INT_SIG_INT = RESP_CTRL1
RESP_INT_SIG_EXT = (RESP_CTRL1 | RESP_CTRL0)

RESP_FREQ2 = 0x80
RESP_FREQ1 = 0x40
RESP_FREQ0 = 0x20
SINGLE_SHOT = 0x08
WCT_TO_RLD = 0x04
PD_LOFF_COMP = 0x02

CONFIG4_const = 0x00

RESP_FREQ_64k_Hz = 0x00
RESP_FREQ_32k_Hz = RESP_FREQ0
RESP_FREQ_16k_Hz = RESP_FREQ1
RESP_FREQ_8k_Hz = (RESP_FREQ1 | RESP_FREQ0)
RESP_FREQ_4k_Hz = RESP_FREQ2
RESP_FREQ_2k_Hz = (RESP_FREQ2 | RESP_FREQ0)
RESP_FREQ_1k_Hz = (RESP_FREQ2 | RESP_FREQ1)
RESP_FREQ_500_Hz = (RESP_FREQ2 | RESP_FREQ1 | RESP_FREQ0)

aVF_CH6 = 0x80
aVL_CH5 = 0x40
aVR_CH7 = 0x20
avR_CH4 = 0x10
PD_WCTA = 0x08
WCTA2 = 0x04
WCTA1 = 0x02
WCTA0 = 0x01

WCT1_const = 0x00

WCTA_CH1P = 0x00
WCTA_CH1N = WCTA0
WCTA_CH2P = WCTA1
WCTA_CH2N = (WCTA1 | WCTA0)
WCTA_CH3P = WCTA2
WCTA_CH3N = (WCTA2 | WCTA0)
WCTA_CH4P = (WCTA2 | WCTA1)
WCTA_CH4N = (WCTA2 | WCTA1 | WCTA0)

PD_WCTC = 0x80
PD_WCTB = 0x40
WCTB2 = 0x20
WCTB1 = 0x10
WCTB0 = 0x08
WCTC2 = 0x04
WCTC1 = 0x02
WCTC0 = 0x01

WCT2_const = 0x00

WCTB_CH1P = 0x00
WCTB_CH1N = WCTB0
WCTB_CH2P = WCTB1
WCTB_CH2N = (WCTB1 | WCTB0)
WCTB_CH3P = WCTB2
WCTB_CH3N = (WCTB2 | WCTB0)
WCTB_CH4P = (WCTB2 | WCTB1)
WCTB_CH4N = (WCTB2 | WCTB1 | WCTB0)

WCTC_CH1P = 0x00
WCTC_CH1N = WCTC0
WCTC_CH2P = WCTC1
WCTC_CH2N = (WCTC1 | WCTC0)
WCTC_CH3P = WCTC2
WCTC_CH3N = (WCTC2 | WCTC0)
WCTC_CH4P = (WCTC2 | WCTC1)
WCTC_CH4N = (WCTC2 | WCTC1 | WCTC0)

MISC1_const = 0
SRB1 = 0x20



#!/usr/bin/env python

# Example program to show how to read a multi-channel time series from LSL.

from pylsl import StreamInlet, resolve_stream

# first resolve an EEG stream on the lab network
print("looking for an EEG stream...")
streams = resolve_stream('type', 'EEG')

# create a new inlet to read from the stream
inlet = StreamInlet(streams[0])

while True:
    # get a new sample (you can also omit the timestamp part if you're not
    # interested in it)
    sample, timestamp = inlet.pull_sample()
    print(timestamp, sample)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import os
import sys
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

about = {}


with open(os.path.join(here, "hackeeg", "__version__.py")) as f:
    exec(f.read(), about)

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

setup(name='hackeeg',
      version=about["__version__"],
      description='Python client library for HackEEG Arduino Due shield',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/starcat-io/hackeeg-client-python',
      author='Starcat LLC',
      author_email='adam@starcat.io',
      license='Apache 2.0',
      packages=['hackeeg'],
      install_requires=[
			"pyserial",
			"bitstring",
			"numpy",
			"jsonlines",
			"msgpack",
			"autopep8",
			"gnureadline",
			"pylsl",
      ],
      scripts=[
            "bin/example.py",
            "bin/hackeeg_shell",
            "bin/hackeeg_shell.py",
            "bin/hackeeg_stream",
            "bin/hackeeg_stream.py",
            "bin/lsl_receive_data.py",
            "bin/simple_lsl_receive_and_plot.py",
      ],
      zip_safe=False)



#!/usr/bin/env python

import numpy as np
from pylsl import StreamInlet, resolve_stream, local_clock
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui


plot_duration = 2.0


# first resolve an EEG stream on the lab network
print("looking for an EEG stream...")
streams = resolve_stream('type', 'EEG')

# create a new inlet to read from the stream
inlet = StreamInlet(streams[0])

# Create the pyqtgraph window
win = pg.GraphicsWindow()
win.setWindowTitle('LSL Plot ' + inlet.info().name())
plt = win.addPlot()
plt.setLimits(xMin=0.0, xMax=plot_duration, yMin=-1.0 * (inlet.channel_count - 1), yMax=1.0)

t0 = [local_clock()] * inlet.channel_count
curves = []
for ch_ix in range(inlet.channel_count):
    curves += [plt.plot()]


def update():
    global inlet, curves, t0
    # Read data from the inlet. Use a timeout of 0.0 so we don't block GUI interaction.
    chunk, timestamps = inlet.pull_chunk(timeout=0.0, max_samples=32)
    if timestamps:
        timestamps = np.asarray(timestamps)
        y = np.asarray(chunk)

        for ch_ix in range(inlet.channel_count):
            old_x, old_y = curves[ch_ix].getData()
            if old_x is not None:
                old_x += t0[ch_ix]  # Undo t0 subtraction
                this_x = np.hstack((old_x, timestamps))
                this_y = np.hstack((old_y, y[:, ch_ix] - ch_ix))
            else:
                this_x = timestamps
                this_y = y[:, ch_ix] - ch_ix
            t0[ch_ix] = this_x[-1] - plot_duration
            this_x -= t0[ch_ix]
            b_keep = this_x >= 0
            curves[ch_ix].setData(this_x[b_keep], this_y[b_keep])



timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(50)


# Start Qt event loop unless running in interactive mode or using pyside.
if __name__ == '__main__':
    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



#!/usr/bin/env python

# Notes:
# To capture at 16,000 samples per second,
# the Arduino Due driver's SPI DMA must be on,
# and the driver communication mode must be set to MessagePack
# (--messagepack option)

import argparse
import threading
import uuid
import time
import sys
import select
import tty
import termios

from pylsl import StreamInfo, StreamOutlet

import hackeeg
from hackeeg import ads1299
from hackeeg.driver import SPEEDS, GAINS, Status

DEFAULT_NUMBER_OF_SAMPLES_TO_CAPTURE = 50000


class HackEegTestApplicationException(Exception):
    pass


class NonBlockingConsole(object):

    def __enter__(self):
        self.old_settings = termios.tcgetattr(sys.stdin)
        tty.setcbreak(sys.stdin.fileno())
        return self

    def __exit__(self, type, value, traceback):
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.old_settings)

    def get_data(self):
        if select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], []):
            return sys.stdin.read(1)
        return False


class HackEegTestApplication:
    """HackEEG commandline tool."""

    def __init__(self):
        self.serial_port_name = None
        self.hackeeg = None
        self.debug = False
        self.channel_test = False
        self.quiet = False
        self.hex = False
        self.messagepack = False
        self.channels = 8
        self.samples_per_second = 500
        self.gain = 1
        self.max_samples = 5000
        self.lsl = False
        self.lsl_info = None
        self.lsl_outlet = None
        self.lsl_stream_name = "HackEEG"
        self.stream_id = str(uuid.uuid4())
        self.read_samples_continuously = True
        self.continuous_mode = False
        self.non_blocking_console = NonBlockingConsole()
        # self.debug = True

    def find_dropped_samples(self, samples, number_of_samples):
        sample_numbers = {self.get_sample_number(sample): 1 for sample in samples}
        correct_sequence = {index: 1 for index in range(0, number_of_samples)}
        missing_samples = [sample_number for sample_number in correct_sequence.keys()
                           if sample_number not in sample_numbers]
        return len(missing_samples)

    def get_sample_number(self, sample):
        sample_number = sample.get('sample_number', -1)
        return sample_number

    def read_keyboard_input(self):
        char = self.non_blocking_console.get_data()
        if char:
            self.read_samples_continuously = False

    def setup(self, samples_per_second=500, gain=1, messagepack=False):
        if samples_per_second not in SPEEDS.keys():
            raise HackEegTestApplicationException("{} is not a valid speed; valid speeds are {}".format(
                samples_per_second, sorted(SPEEDS.keys())))
        if gain not in GAINS.keys():
            raise HackEegTestApplicationException("{} is not a valid gain; valid gains are {}".format(
                gain, sorted(GAINS.keys())))

        self.hackeeg.stop_and_sdatac_messagepack()
        self.hackeeg.sdatac()
        self.hackeeg.blink_board_led()
        sample_mode = SPEEDS[samples_per_second] | ads1299.CONFIG1_const
        self.hackeeg.wreg(ads1299.CONFIG1, sample_mode)

        gain_setting = GAINS[gain]

        self.hackeeg.disable_all_channels()
        if self.channel_test:
            self.channel_config_test()
        else:
            self.channel_config_input(gain_setting)


        # Route reference electrode to SRB1: JP8:1-2, JP7:NC (not connected)
        # use this with humans to reduce noise
        self.hackeeg.wreg(ads1299.MISC1, ads1299.SRB1 | ads1299.MISC1_const)

        # Single-ended mode - setting SRB1 bit sends mid-supply voltage to the N inputs
        # use this with a signal generator
        # self.hackeeg.wreg(ads1299.MISC1, ads1299.SRB1)

        # Dual-ended mode
        self.hackeeg.wreg(ads1299.MISC1, ads1299.MISC1_const)
        # add channels into bias generation
        # self.hackeeg.wreg(ads1299.BIAS_SENSP, ads1299.BIAS8P)

        if messagepack:
            self.hackeeg.messagepack_mode()
        else:
            self.hackeeg.jsonlines_mode()
        self.hackeeg.start()
        self.hackeeg.rdatac()
        return

    def channel_config_input(self, gain_setting):
        # all channels enabled
        # for channel in range(1, 9):
        #     self.hackeeg.wreg(ads1299.CHnSET + channel, ads1299.TEST_SIGNAL | gain_setting )

        # self.hackeeg.wreg(ads1299.CHnSET + 1, ads1299.INT_TEST_DC | gain_setting)
        # self.hackeeg.wreg(ads1299.CHnSET + 6, ads1299.INT_TEST_DC | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 1, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 2, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 3, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 4, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 5, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 6, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 7, ads1299.ELECTRODE_INPUT | gain_setting)
        self.hackeeg.wreg(ads1299.CHnSET + 8, ads1299.ELECTRODE_INPUT | gain_setting)

    def channel_config_test(self):
        # test_signal_mode = ads1299.INT_TEST_DC | ads1299.CONFIG2_const
        test_signal_mode = ads1299.INT_TEST_4HZ | ads1299.CONFIG2_const
        self.hackeeg.wreg(ads1299.CONFIG2, test_signal_mode)
        self.hackeeg.wreg(ads1299.CHnSET + 1, ads1299.INT_TEST_DC | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 2, ads1299.SHORTED | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 3, ads1299.MVDD | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 4, ads1299.BIAS_DRN | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 5, ads1299.BIAS_DRP | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 6, ads1299.TEMP | ads1299.GAIN_1X)
        self.hackeeg.wreg(ads1299.CHnSET + 7, ads1299.TEST_SIGNAL | ads1299.GAIN_1X)
        self.hackeeg.disable_channel(8)

        # all channels enabled
        # for channel in range(1, 9):
        #     self.hackeeg.wreg(ads1299.CHnSET + channel, ads1299.TEST_SIGNAL | gain_setting )
        pass



    def parse_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("serial_port", help="serial port device path",
                            type=str)
        parser.add_argument("--debug", "-d", help="enable debugging output",
                            action="store_true")
        parser.add_argument("--samples", "-S", help="how many samples to capture",
                            default=DEFAULT_NUMBER_OF_SAMPLES_TO_CAPTURE, type=int)
        parser.add_argument("--continuous", "-C", help="read data continuously (until <return> key is pressed)",
                            action="store_true")
        parser.add_argument("--sps", "-s",
                            help=f"ADS1299 samples per second setting- must be one of {sorted(list(SPEEDS.keys()))}, default is {self.samples_per_second}",
                            default=self.samples_per_second, type=int)
        parser.add_argument("--gain", "-g",
                            help=f"ADS1299 gain setting for all channels– must be one of {sorted(list(GAINS.keys()))}, default is {self.gain}",
                            default=self.gain, type=int)
        parser.add_argument("--lsl", "-L",
                            help=f"Send samples to an LSL stream instead of terminal",
                            action="store_true"),
        parser.add_argument("--lsl-stream-name", "-N",
                            help=f"Name of LSL stream to create",
                            default=self.lsl_stream_name, type=str),
        parser.add_argument("--messagepack", "-M",
                            help=f"MessagePack mode– use MessagePack format to send sample data to the host, rather than JSON Lines",
                            action="store_true")
        parser.add_argument("--channel-test", "-T",
                            help=f"set the channels to internal test settings for software testing",
                            action="store_true")
        parser.add_argument("--hex", "-H",
                            help=f"hex mode– output sample data in hexidecimal format for debugging",
                            action="store_true")
        parser.add_argument("--quiet", "-q",
                            help=f"quiet mode– do not print sample data (used for performance testing)",
                            action="store_true")
        args = parser.parse_args()
        if args.debug:
            self.debug = True
            print("debug mode on")
        self.samples_per_second = args.sps
        self.gain = args.gain

        if args.continuous:
            self.continuous_mode = True

        if args.lsl:
            self.lsl = True
            if args.lsl_stream_name:
                self.lsl_stream_name = args.lsl_stream_name
            self.lsl_info = StreamInfo(self.lsl_stream_name, 'EEG', self.channels, self.samples_per_second, 'int32',
                                       self.stream_id)
            self.lsl_outlet = StreamOutlet(self.lsl_info)

        self.serial_port_name = args.serial_port
        self.hackeeg = hackeeg.HackEEGBoard(self.serial_port_name, baudrate=2000000, debug=self.debug)
        self.max_samples = args.samples
        self.channel_test = args.channel_test
        self.quiet = args.quiet
        self.hex = args.hex
        self.messagepack = args.messagepack
        self.hackeeg.connect()
        self.setup(samples_per_second=self.samples_per_second, gain=self.gain, messagepack=self.messagepack)

    def process_sample(self, result, samples):
        if result:
            status_code = result.get(self.hackeeg.MpStatusCodeKey)
            data = result.get(self.hackeeg.MpDataKey)
            samples.append(result)
            if status_code == Status.Ok and data:
                if not self.quiet:
                    timestamp = result.get('timestamp')
                    sample_number = result.get('sample_number')
                    ads_gpio = result.get('ads_gpio')
                    loff_statp = result.get('loff_statp')
                    loff_statn = result.get('loff_statn')
                    channel_data = result.get('channel_data')
                    data_hex = result.get('data_hex')
                    print(
                        f"timestamp:{timestamp} sample_number: {sample_number}| gpio:{ads_gpio} loff_statp:{loff_statp} loff_statn:{loff_statn}   ",
                        end='')
                    if self.hex:
                        print(data_hex)
                    else:
                        for channel_number, sample in enumerate(channel_data):
                            print(f"{channel_number + 1}:{sample} ", end='')
                        print()
                if self.lsl:
                    self.lsl_outlet.push_sample(channel_data)
            else:
                if not self.quiet:
                    print(data)
        else:
            print("no data to decode")
            print(f"result: {result}")

    def main(self):
        self.parse_args()

        samples = []
        sample_counter = 0

        end_time = time.perf_counter()
        start_time = time.perf_counter()
        while ((sample_counter < self.max_samples and not self.continuous_mode) or \
               (self.read_samples_continuously and self.continuous_mode)):
            result = self.hackeeg.read_rdatac_response()
            end_time = time.perf_counter()
            sample_counter += 1
            if self.continuous_mode:
                self.read_keyboard_input()
            self.process_sample(result, samples)

        duration = end_time - start_time
        self.hackeeg.stop_and_sdatac_messagepack()
        self.hackeeg.blink_board_led()

        print(f"duration in seconds: {duration}")
        samples_per_second = sample_counter / duration
        print(f"samples per second: {samples_per_second}")
        dropped_samples = self.find_dropped_samples(samples, sample_counter)
        print(f"dropped samples: {dropped_samples}")


if __name__ == "__main__":
    HackEegTestApplication().main()



#!/usr/bin/env python

import argparse
import cmd
import sys
import time

import hackeeg
from hackeeg import ads1299


# TODO
# - argparse commandline arguments
# - JSONLines
# -MessagePack

class HackEEGShell(cmd.Cmd):
    intro = 'Welcome to the HackEEG shell. Type help or ? to list commands.\n'
    prompt = '(hackeeg) '
    file = None

    def __init__(self):
        super().__init__()
        self.serial_port_name = None
        self.hackeeg = None
        self.debug = False
        self.hex = True

    def _format_response(self, response):
        if self.debug:
            print(f"_format_response: {response}")
        if self.hackeeg.mode == self.hackeeg.JsonLinesMode:
            status_code = response.get(self.hackeeg.StatusCodeKey)
            status_text = response.get(self.hackeeg.StatusTextKey)
            data = response.get(self.hackeeg.DataKey)
        elif self.hackeeg.mode == self.hackeeg.MessagePackMode:
            status_code = response.get(self.hackeeg.MpStatusCodeKey)
            status_text = response.get(self.hackeeg.MpStatusTextKey)
            data = response.get(self.hackeeg.MpDataKey)
        if status_code and status_code == hackeeg.Status.Ok:
            print("Ok", end='')
            if data:
                print(f"  Data: ", end='')
                if isinstance(data, list):
                    self._format_list(data)
                elif isinstance(data, int):
                    self._format_int(data)
                else:
                    print(f"{data}", end='')
            print()
        else:
            print(f"Error: {response}")

    def _format_list(self, data):
        if len(data) == 0:
            return
        for item in data:
            self._format_int(data)

    def _format_int(self, item):
        if self.hex:
            print(f"{item:02x} ", end='')
        else:
            print(f"{item:02d} ", end='')
        print()

    def do_debug(self, arg):
        """Enables debugging output in the client software. Usage: debug <on|off>"""
        usage = "Usage: debug [on|off]"
        if arg == "on":
            self.debug = True
        elif arg == "off":
            self.debug = False
        else:
            print(f"debug flag is currently set to {self.debug}.")
        self.hackeeg.set_debug(self.debug)
        return

    def do_hex(self, arg):
        """Sets numerical output format to hexidecimal or decimal. Defaults to hexidecimal. Usage: hex <on|off>"""
        usage = "Usage: hexmode [on|off]"
        if arg == "on":
            self.hex = True
        elif arg == "off":
            self.hex = False
        print(self._hex_mode_string())
        return

    def _hex_mode_string(self):
        return f"Numerical output is currently set to {'hexidecimal' if self.hex else 'decimal'}."

    def do_nop(self, arg):
        """No operation – does nothing."""
        self._format_response(self.hackeeg.nop())

    def do_version(self, arg):
        """Returns HackEEG driver version number."""
        result = self.hackeeg.version()
        print(result[self.hackeeg.StatusTextKey])
        self._format_response(result)

    def do_status(self, arg):
        """Returns HackEEG driver status."""
        self._format_response(self.hackeeg.status())

    def do_text(self, arg):
        """Sets driver communication protocol to text and exits the HackEEG command shell. 
           Useful if you want to communicate directly with the driver using text mode."""
        self._format_response(self.hackeeg.text_mode())

    def do_jsonlines(self, arg):
        """Sets driver communication protocol to jsonlines. (Not implemented yet.)"""
        self._format_response(self.hackeeg.jsonlines_mode())

    def do_messagepack(self, arg):
        """Sets driver communication protocol to MessagePack. (Not implemented yet.)"""
        self._format_response(self.hackeeg.messagepack_mode())

    def do_micros(self, arg):
        """Returns microseconds since the Arduino booted up."""
        self._format_response(self.hackeeg.micros())

    def do_serialnumber(self, arg):
        """Returns board serial number. (Not implemented yet.)"""
        self._format_response(self.hackeeg.serialnumber())

    def do_boardledon(self, arg):
        """Turns the HackEEG board LED on."""
        self._format_response(self.hackeeg.boardledon())

    def do_boardledoff(self, arg):
        """Turns the HackEEG board LED off."""
        self._format_response(self.hackeeg.boardledoff())

    def do_ledon(self, arg):
        """Turns the Arduino LED on."""
        self._format_response(self.hackeeg.ledon())

    def do_ledoff(self, arg):
        """Turns the Arduino LED off."""
        self._format_response(self.hackeeg.ledoff())

    def do_blink(self, arg):
        """Turns the HackEEG board LED off, then on, then off."""
        self.hackeeg.boardledoff()
        self.hackeeg.boardledon()
        time.sleep(0.2)
        self._format_response(self.hackeeg.boardledoff())

    def do_rreg(self, arg):
        """Reads an ADS1299 register and returns the value."""
        usage = "rreg [register_number]"
        arglist = []
        try:
            arglist = parse_registers(arg)
        except HackEEGArgumentException as e:
            pass
        if len(arglist) == 0:
            print("No register number given.")
            print(usage)
        elif len(arglist) > 1:
            print("Too many arguments.")
            print(usage)
        else:
            self._format_response(self.hackeeg.rreg(arglist[0]))

    def do_wreg(self, arg):
        """Writees an ADS1299 register."""
        arglist = parse_registers(arg)
        if len(arglist) == 0:
            print("No register number given.")
        elif len(arglist) == 1:
            print("No register value given.")
        elif len(arglist) > 2:
            print("Too many arguments.")
        else:
            self._format_response(self.hackeeg.wreg(arglist[0], arglist[1]))

    def do_wakeup(self, arg):
        """Sends the WAKEUP command to the ADS1299. (Not implemented yet.)"""
        self._format_response(self.hackeeg.nop())

    def do_standby(self, arg):
        """Sends the STANDBY command to the ADS1299. (Not implemented yet.)"""
        self._format_response(self.hackeeg.nop())

    def do_reset(self, arg):
        """Sends the RESET command to the ADS1299."""
        self._format_response(self.hackeeg.reset())

    def do_start(self, arg):
        """Sends the START command to the ADS1299. (Not implemented yet.)"""
        self._format_response(self.hackeeg.start())

    def do_stop(self, arg):
        """Sends the STOP command to the ADS1299."""
        self._format_response(self.hackeeg.stop())

    def do_rdatac(self, arg):
        """Sends the RDATAC command to the ADS1299."""
        result = self.hackeeg.rdatac()
        if self.hackeeg.ok(result):
            print("rdatac mode on.")
        else:
            print(f"rdatac failed to turn on: {result}")
        self._format_response(result)

    def do_sdatac(self, arg):
        """Sends the SDATAC command to the ADS1299."""
        self._format_response(self.hackeeg.sdatac())

    def do_rdata(self, arg):
        """Sends the RDATA command to the ADS1299. (Not implemented yet.)"""
        self._format_response(self.hackeeg.nop())

    def do_getdata(self, arg):
        """Gets data from the HackEEG driver's ringbuffer. (Not implemented yet.)"""
        self._format_response(self.hackeeg.nop())

    def do_setup(self, arg):
        """Sets up the ADS1299."""
        self.hackeeg.blink_board_led()
        self.hackeeg.reset()
        sample_mode = ads1299.HIGH_RES_250_SPS | ads1299.CONFIG1_const
        self.hackeeg.wreg(ads1299.CONFIG1, sample_mode)
        self.hackeeg.disable_channel(1)
        self.hackeeg.disable_channel(2)
        self.hackeeg.disable_channel(3)
        self.hackeeg.disable_channel(4)
        #self.hackeeg.disable_channel(5)
        self.hackeeg.disable_channel(6)
        #self.hackeeg.disable_channel(7)
        self.hackeeg.disable_channel(8)
        self.hackeeg.wreg(ads1299.CH5SET, ads1299.TEST_SIGNAL | ads1299.GAIN_2X)
        #self.hackeeg.wreg(ads1299.CH7SET, ads1299.TEMP | ads1299.GAIN_12X)
        self.hackeeg.rreg(ads1299.CH5SET)
        #self.hackeeg.wreg(ads1299.CH8SET, ads1299.ELECTRODE_INPUT | ads1299.GAIN_24X)
        self.hackeeg.wreg(ads1299.CH7SET, ads1299.ELECTRODE_INPUT | ads1299.GAIN_2X)
        #self.hackeeg.wreg(ads1299.CH2SET, ads1299.ELECTRODE_INPUT | ads1299.GAIN_24X)
        self.hackeeg.rreg(ads1299.CH8SET)
        #self.hackeeg.rreg(ads1299.CH2SET)
        self.hackeeg.wreg(ads1299.CH2SET, ads1299.ELECTRODE_INPUT | ads1299.GAIN_24X)
        # Unipolar mode - setting SRB1 bit sends mid-supply voltage to the N inputs
        self.hackeeg.wreg(ads1299.MISC1, ads1299.SRB1)
        # add channels into bias generation
        self.hackeeg.wreg(ads1299.BIAS_SENSP, ads1299.BIAS8P)
        self.hackeeg.rdatac()
        return

    def do_exit(self, arg):
        """Exit the HackEEG commandline."""
        sys.exit(0)

    def do_quit(self, arg):
        """Exit the HackEEG commandline."""
        sys.exit(0)

    def default(self, line):
        if line == 'EOF':
            sys.exit(0)
        print("Unrecognized command.")

    def setup(self, samplesPerSecond=500):
        if samplesPerSecond not in hackeeg.SPEEDS.keys():
            raise hackeeg.HackEEGException("{} is not a valid speed; valid speeds are {}".format(
                samplesPerSecond, sorted(hackeeg.SPEEDS.keys())))
        self.hackeeg.jsonlines_mode()
        # self.hackeeg.sdatac()
        time.sleep(1)
        # self.hackeeg.send("reset")
        self.hackeeg.blink_board_led()
        return

    def main(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("serial_port", help="serial port device path",
                            type=str)
        parser.add_argument("--debug", "-d", help="enable debugging output",
                            action="store_true")
        args = parser.parse_args()
        if args.debug:
            self.debug = True
            print("debug mode on")
        self.serial_port_name = args.serial_port
        self.hackeeg = hackeeg.HackEEGBoard(self.serial_port_name, debug=self.debug)
        self.hackeeg.connect()
        intro_message = "Welcome to the HackEEG shell. Type help or ? to list commands.\n"
        intro_message += self._hex_mode_string() + '\n'
        self.cmdloop(intro_message)

class HackEEGArgumentException(Exception):
    pass

def parse_registers(arg):
    """Convert a series of zero or more numbers to an argument tuple"""
    result = tuple(map(int, arg.split()))
    for item in result:
        if item > 255 or item < 0:
            raise HackEEGArgumentException("Register and value must be an integer between 0 and 255.")
    return result


if __name__ == "__main__":
    HackEEGShell().main()
